#define VERSION "2.01 [24-Jun-1995]"

/*
 **********************************************************************
 *                                                                    *
 * dired [][dir][files]  Stuart Cracraft (mclure@sri-unix) Sept 1980  *
 *								      *
 *      Directory editor.					      *
 *		edit/delete files in a directory or a file list	      *
 *              compile with the berkeley termlib archive             *
 *              and don't forget to change the 'helpfile' and         *
 *              'dirednam' strings to reflect your own setup.         *
 *                                                                    *
 *      Note: if you make improvements, I'd like to get them too.     *
 *                      Stuart Cracraft mclure@sri-unix,              *
 *				ucbvax!menlo70!sri-unix!mclure	      *
 *	so would I:	Jay Lepreau 	lepreau@utah-20,	      *
 *			decvax!{harpo,randvax}!utah-cs!lepreau	      *
 *      me too:         Charles Hill     philabs!crh                  *
 **********************************************************************
*/

/*
 *	Enhanced by J.Lepreau, Univ of Utah, 8-10/81:
 *	      --bunch of stuff, including more cmds, 2 window mode,
 *		bufering output, pathname substition in !escapes,
 *		initial sort options, ^G escape from sorts, ^U escape
 *		from !escapes, empty dir check.
 *


Modified 12/81 by Lepreau:
1. Fix up aborting of command escapes, with mildy tricky re-display.
2. Make default window size be half-screen (2 window mode).
   Added option -wf to get full screen.
3. Add check for too many files to avoid core-dump.
4. Fix bug in cmd escapes: must reset caught signals before invocation.

Modified 6/82 by Charles Hill (crh)
		 Philips Labs
		 345 Scarborough Rd.
                 Briarcliff Manor, N.Y. 10510
		 philabs!crh

1. Eliminated need for terminal home capability. (Many termcap entries
   leave this out when there is cm). Changed body of home ()
   routine to a call to curxy(0,0).
2. Made default editor "vi".
3. On exit, screen is not blanked; cursor moves to last line. Line is
   cleared
4. In "type", only outputs 79 chars/line so terms with automatic margins
   will work ok.
5. Changed type () to handle tabs better.
6. Changed ceol to work for terms w/o clear to eoln. Single call to ceod
   eliminated. If no clear to eod, this is simulated. But it's better to
   use full screen on such terminals.
7. Changed type () to check if a file is a text file first. On some
   terminals, outputting arbitrary binary can put them in a weird mode,
   and screw up the display.

Additional enhancements 8/82 crh
1. Added pattern matching in the form of /, n, and N commands.
2. Added toggle split-screen command \.
3. Added p command to print path name.
4. Change to "c" shell for ! command.
5. Fixed bug for single file arg when not a dir.
6. Added 'A', 'Q' commands and "linear window indicator".
7. Added counts to <up>, <down>, f, and b commands.
8. Added "G" as in "vi".
9. Added "E" command to go up a dir, and code to print node name.
10.Added "R" command to re-read dir.

****** 4.2 version 8/83 Charles Hill (crh) *****

The 4.2 version is shorter because 4.2 has system routines opendir,
  readir, etc. and -r option of rm

fixes 1/84 crh
1) show embedded control chars
2) eliminate 1 line windows except where necessary
3) cursor bug on deleting last file

*********************************

To compile: cc -o dired dired.c cshsystem.o -ltermcap /a/gam/lib/lib.a
                                                        ^ for getshell()
*/


/*
   things to consider
   -    perhaps should be able to edit protection field and
		be able to change protection instantly at the
		touch of a key (or queue the change for later exit)

   --	Would be nice if there were cmds (e.g. 'c') which would update
		1) the file-info & re-display it, & 2) update whole
		display's info.
*/

#if defined(__STDC__) || defined(__cplusplus)
#define STDC	1
#define ARGS(s) s
#else
#define STDC	0
#define ARGS(s) ()
#define const
#endif

#ifdef vax				/* this does for now */
#define VFORK
#endif

#ifdef V4P2				/* crh 3/84 */
#ifndef VFORK
#define VFORK
#endif
#endif

#define	False		0
#define	True		1
#define	NOTDELETED	0
#define	DELETED		1

/*	Sort Orders */
#define NAME		0
#define	SIZE		1
#define	WRITE		2
#define	READ		3
#define CREATE		4
const char sortstr[] = "nswrc";	/* must be in order by above */

#define ESC         '\033'
#define CURSOR          48	/* X-coord of cursor, just b4 file */

#define	eq(a, b)	(strcmp(a, b) == 0)

#include <stdio.h>
#include <sys/types.h>
#include <string.h>

#include <signal.h>

#if STDC
#define HAVE_VOID_HANDLER	1
#endif

#ifndef SIGTSTP
#define SIGTSTP 18
#endif

#include <time.h>
#include <sys/wait.h>			/* for wait() et al prototypes */

#if defined(__NeXT__)
#define wait_arg_t	union wait
#else
#define wait_arg_t	int
#endif

#if STDC
#include <stdlib.h>
#endif

#if defined(__host_mips) && defined(_SYSTYPE_SVR4)
	/* MIPS RC/6280 RISCos 2.1.1 -- no Standard C compiler yet */
#include <stdlib.h>
#endif

#if !defined(__NeXT__)
#include <unistd.h>			/* for read() et al prototypes */
#endif

#ifdef ardent
#include <sys/43ioctl.h>		/* for TIOxxxxx */
#include <sys/sysmacros.h>		/* for major(), minor() prototypes */
#else
#include <fcntl.h>
#endif

#if defined(HAVE_TERMIO_H)
#include <termio.h>			/* needed on Sun Solaris 2.3 */
#endif

#if defined(HAVE_TERMIOS_H)
#include <termios.h>
static struct termios tty_save;	/* XPG3, POSIX.1, FIPS 151-1 interface */
#else /* NOT defined(HAVE_TERMIOS_H) */
#include <sgtty.h>
static struct sgttyb   ioctlb;
#endif /* defined(HAVE_TERMIOS_H) */

#ifdef __sgi
#include <sys/sysmacros.h>		/* for major(), minor() prototypes */
#endif

#if defined(HAVE_MKDEV_H)
#include <sys/mkdev.h>			/* for major(), minor() prototypes */
#endif

#if !defined(HAVE_GETWD)
#undef getwd
#define getwd(p)	getcwd(p,sizeof(p))
#endif

#include <curses.h>

#if HAVE_TERM_H
#include <term.h>			/* for tgetent() et al prototypes */
#endif

#include <sys/param.h>
#include <sys/stat.h>

#if HAVE_DIRENT_H
#include <dirent.h>
#define D_NAMLEN(dp)		(int)strlen((dp)->d_name)
#define HAVE_OPENDIR		1
#else
#include <sys/dir.h>
#define dirent direct			/* we use dirent in rest of code */
#if defined(HAVE_SYSTEM7_DIR_H)
/* I don't think any systems with <sys/dir.h> but no d_namlen entry
exist any more, but if you have one, this should do the trick. */
#define D_NAMLEN(dp)	14		/* really old UNIX System 7 style: */
					/* d_name[] that might lack a NUL */
#else
#define D_NAMLEN(dp)	(int)((dp)->d_namlen) /* newer <sys/dir.h> style */
					/* contains explicit length field */
#endif
#endif /* HAVE_DIRENT_H */

#define DIRPBLOCK	(BUFSIZ / sizeof(struct dirent))
				/* number of directory entries per block */

/* Supply prototypes for library functions that are missing on some systems */

#if defined(__sgi)
extern char*	re_comp ARGS((const char *s));
extern int	re_exec ARGS((const char *s));
#else
extern char*	re_comp ARGS((char *s));
extern int	re_exec ARGS((char *s));
#endif

#if defined(__alpha)
extern char*	tgetstr ARGS((const char *s, char **p));
#endif

#if defined(__NeXT__)
extern char*	tgetent ARGS((char *s, char *name));
extern int	tgetnum ARGS((const char id[2]));
extern char*	tgetstr ARGS((const char *s, char **p));
extern char*	tgoto ARGS((const char *s, int column, int line));
extern int	tputs ARGS((char *str, int affcnt, int (*putc_)(char)));
#endif

#if defined(__sgi)
extern char*	tgoto ARGS((const char *s, int column, int line));
#endif

#if defined(sun)
#if !defined(__solaris__)
extern char*	getenv ARGS((const char *name));
extern char*	tgetstr ARGS((const char *s, char **p));
#endif
extern char*	tgoto ARGS((const char *s, int column, int line));
#endif

#if defined(ultrix)
extern char*	getenv ARGS((const char *name));
extern char*	tgetstr ARGS((const char *s, char **p));
#endif

#ifdef CTRL
#undef CTRL
#endif
#define CTRL(c)		(c & 037)


/****** local system dependent */
#ifndef DIREDFILE
#define DIREDFILE	"/usr/local/bin/dired"
#endif

#ifndef HELPFILE
#define HELPFILE	"/usr/local/bin/dired.hlp"
#endif

static const char   *dirednam = DIREDFILE; /* Where dired is */
static const char   *helpfile = HELPFILE; /* Where helpfile lives */

#define MAXFILEPATHLEN	1024		/* longest file path name on any */
					/* current UNIX system */

#define MAXLBUFF MAXFILEPATHLEN

#define MAXWIDTH 255			/* widest lines we will use */

static char   divider[MAXWIDTH+1], newdivider[MAXWIDTH+1];
					/* divides the windows */
static int    LW;			/* the last pos of the linear window */
#define DIVCHAR	'-'			/* makes up the divider */

#ifndef MOREPGM				/* can override at compile time */
#if defined(__convex__)
#define MOREPGM	"/usr/ucb/more"		/* program to page thru a file */
#else
#define MOREPGM	"/usr/local/bin/less"	/* program to page thru a file */
#endif /* MOREPGM */
#endif /* defined(__convex__) */

#ifndef MAXFILES /* This way can override on the cmd line for a huge version */
#define MAXFILES	10240
#endif

#define ISDIGIT(C)	(C <= '9') && (C >= '0')
#define ABORTSTAT	69

#ifdef V4P2				/* long file names supported */
#ifndef MAXNAMLEN
#define MAXNAMLEN 255
#endif
#define MAXDNAME MAXNAMLEN + 1
#else
#define MAXDNAME 30
#endif /* V4P2 */

struct lbuf
{
    unsigned long	lnum;
    unsigned long	lnl;
    long		lsize;
    long		latime;
    long		lctime;
    long		lmtime;
    int			ltype;
    short   		deleted;
    short		lflags;
    short		luid;
    short		lgid;
    union
    {
	char		lname[MAXDNAME];  /*****crh 8/83  ****/
	char   		*namep;
    } ln;
};

static struct lbuf file[MAXFILES];

static struct stat	statbuf;

#define	ISARG	0100000

#define FNAME(N)	(arglist ? file[N].ln.namep : file[N].ln.lname)
static int	gflg,
	        iflg,
	        lflg,
	        sflg;				/* Random flags */
static int	splitflg;		/* Split screen? */
static int	sortyp;			/* Key to sort on */
static int	errcode = 0;		/* Error variable used by rm */
static int	rflg = 1;		/* Reverse sort flag */
static int	totfiles = 0;		/* Total files */
static int	flags;			/* Gets flags */
static int	blurb;			/* 1=>thing in echo area,0 otherwise  */
static int	numdeleted;		/* Number of files marked deleted */
static long	ttime;			/* Temp time variable */
static long	year;			/* Six months ago */
static long	totblocks = 0;		/* Total blocks */
static int	lastuid = -1;		/* Last uid/gid we handled */
static char	tmpbuf[MAXLBUFF];	/* Random temporary buffer */
static char	tmpbuf2[MAXLBUFF];
static char	userbuf[1024];		/* Temporary buffer for user name */
		/* must be big enough to hold longest LINE in /etc/passwd */
static FILE 	*pwdf;			/* Password/group file and directory files */

static int	curfile = 0;		/* Current file */
static int	topfile;        	/* File at top of screen */
static int	curline = 0;		/* Line that we're on */
static int	scrlen = 999;		/* Default length of dired index part of screen: */
		/* 999 ==> 2 windows (half size), 0 ==> 1 window (full size) */
static int	Worklen = 0;	/* Length of 'working window', the other part*/
static int	Worktop = 0;	/* Top of    " " */
static int	Tscrlen;		/*Total length of screen,minus 1 for cmd line*/
static int	scrwid = 79;		/* Width of screen */

/* NB: Do NOT make CL...PC static, becuase this conflicts with declarations
in DEC Alpha <curses.h>! */
char	   	*CL,			/* Termcap strings we'll use */
		*UP,
		*HO,
		*CM,
		*CD,			/* clear to end of display */
		*CE,
		*AL;			/* insert line */

char		PC;

static char	tbuf[1024],
	        tcapbuf[256];
static int	sigint;
static int	command;		/* Holds last command */
static int	intype=0;

#if defined(HAVE_VOID_HANDLER)
#define handler_t void
#else
#define handler_t int
#endif

int		main ARGS((int argc, char *argv[]));

extern int	cshsystem ARGS((char *s));

static void	blank ARGS((void));
static char*	bldnam ARGS((char *str, int narg, int filidx, char *argv[]));
static void	bscreen ARGS((int count));
static char*	catargs ARGS((char *str, char *argv[]));
static handler_t catchint ARGS((int sig));
static void	ceol ARGS((int x, int y));
static int	compar ARGS((const void *pp1, const void *pp2));
static void	curxy ARGS((int col, int lin));
static void	dcanon ARGS((char *cp));
static void	deletefiles ARGS((int argc, char *argv[]));
static void	dispfile ARGS((int fnum));
static void	dispwin ARGS((int filenum));
static void	downline ARGS((void));
static void	expandpath ARGS((char pname[]));
static void	findfile ARGS((int down));
static int	findfname ARGS((char filename[], int *num));
static void	formstartarg ARGS((char *name));
static void	fscreen ARGS((int count));
static void	Getcap ARGS((void));
static int	Getch ARGS((FILE *fd));
static void	getcountarg ARGS((void));
static void	getfiles ARGS((int argc, char *argv[]));
static char*	getfullpath ARGS((const char *start));
static int	getline ARGS((char s[]));
static int	getname ARGS((int uid, char buf[]));
static int	gstat ARGS((const char *gname));
static void	home ARGS((void));
static void	insline ARGS((void));
static void	locstart ARGS((char strtfile[]));
static char*	makename ARGS((const char *dir, const char *filen, int len));
static int	match ARGS((int filenum));
static long	nblock ARGS((long size));
static int	nottext ARGS((const char fn[]));
static handler_t onctlz ARGS((int sig));

#if defined(__sgi)
static int	outch ARGS((int c));
#else
static int	outch ARGS((char c));
#endif

static void	overflow ARGS((void));
static void	pentry ARGS((int whichone));
static void	pmode ARGS((int aflag));
static void	printpath ARGS((char str[], int printbottom));
static void	prtsplitstr ARGS((const char str[]));
static void	putpad ARGS((char *str));

static void	Readdir ARGS((const char *dir));

static void	rm ARGS((char arg[], int level));

#ifndef V4P2
static int	Rmdir ARGS((char *f));
#endif

static void	Select ARGS((int *pairp));
static void	setdpy ARGS((void));
static void	shortcountmess ARGS((void));
static void	showscreen ARGS((void));
static char*	skipto ARGS((const char *string, const char *charset));
static void	telluser ARGS((const char *msg, const char *arg));
static int	type ARGS((const char *filestr, int waitflg));
static void	typefiles ARGS((void));
static void	typein ARGS((const char *argv));
static void	unsetdpy ARGS((void));
static int	waitchk ARGS((int waitflg));
static void	windsize ARGS((void));

#if 0					/* unused functions */
static void	ceod ARGS((void));
static int	dotname ARGS((char *s));
static void	prname ARGS((void));
static void	upline ARGS((void));
#endif

static int asswin = -1;
static int arglist=0;
static char fullpath[MAXLBUFF];
static char startfile[MAXLBUFF], startarg[MAXLBUFF];
static int  wasstartfile = 0;
static int quitok = 1;
static int countarg;
static int camedown = 0;
static int     wasspecial;
static int     specialcc;
static int	numarg;
static char    nambuf[MAXLBUFF];
static char bufout[BUFSIZ];

#if !defined(MAX_MARKS)
#define MAX_MARKS	1024
#endif

#if !defined(MAX_VIEW)
#define MAX_VIEW	10240
#endif

static int     view_stack[MAX_VIEW];
static int     view_top;


int
main (argc, argv)
int     argc;
char   *argv[];
{
    char	*cp;
    int		diredstat;
    register int     i,
    	    		cc;
    int	    special;		/* flag says had % or # in cmd */
    static char combuf[MAXLBUFF] = " "; /* Holds last ! command */
    wait_arg_t status;
    char   **oldargv;
    const char *ct;
    register char   *t,
    	   	    *tt;
    int     mark_stack[MAX_MARKS];
    int     mark_top;

    mark_top = 0;
    view_top = 0;
    re_comp("/");	/* create an unmatchable initial file pattern */

 /* Get terminal type */

    setbuf(stdout, NULL);	/* for v7 speed up, vax compatibility -fjl */
    Getcap ();
    Tscrlen--;	/* don't use final line, in case it is terminal status
		   line (e.g. line 25 on VT100 display) */

 /* Process arg flags. They must be first! */
    for (i=1; i<argc; i++) {
	if (argv[i][0] != '-')
	    break;
	switch (cc = argv[i][1]) {
	  case 'v':			/* program version */
	    (void)fprintf(stderr,"%s Version %s\n", argv[0], VERSION);
	    exit(0);
	    break;
	  case 'w':			/* Window size */
	    if (argv[i][2] == 'h')
		scrlen = 999;		/* Half */
	    else if (argv[i][2] == 'f')
		scrlen = 0;		/* Full */
	    else {
		scrlen = atoi(&argv[i][2]);
		asswin = scrlen;
	    }
	    break;
	  case 's':			/* Initial sort order */
	  case 'r':
	    sortyp = strchr(sortstr, argv[i][2]) - sortstr;
	    rflg = (cc == 's') ? 1 : -1;
	    break;

	  case 'd': camedown = True;
		    break;
	  case 'F':
	  case 'f':   /* get the start file */
		    strcpy(startfile,&argv[i][2]);
		    wasstartfile=True;
		    if (cc == 'F') quitok=False;
		    break;
	  default:
	    fprintf(stderr, "Unknown option %s, ignored.\n", argv[i]);
	    break;
	 }
    }
    argc -= (i - 1);
    oldargv = argv;
    argv += (i - 1);

    windsize ();                 /* replaced by proc call 8/82 crh */

#if 0
    tt = divider + scrwid + 1;
#else
    tt = divider + scrwid;		/* fix off-by-one error */
#endif
    LW = scrwid;   /* last position in the linear window */
    for (t = divider; t < tt;)	/* arbitrary length */
    	*t++ = DIVCHAR;
    *t = '\0';

    signal (SIGINT, SIG_IGN);
    signal (SIGQUIT, SIG_IGN);

    setdpy ();
    printf ("Reading");

    time (&ttime);
    year = ttime - 6L * 30L * 24L * 60L * 60L;/* 6 months ago */

    lflg = 1;
    gflg = iflg = sflg = 0;
    if (lflg)
    {
	ct = "/etc/passwd";
	if (gflg)
	    ct = "/etc/group";
	pwdf = fopen (ct, "r");
    }


    numarg = argc;
    getfiles (argc,argv);

    /* extract the full path name */
    if (!arglist) {
      if (argc==1) cp=getfullpath(".");
      else if (argv[1][0] != '/')  cp=getfullpath(argv[1]);
      else cp=argv[1];
      dcanon(cp);
      strcpy(fullpath,cp);
    }

    qsort (file, totfiles, sizeof (struct lbuf), compar);
    if (wasstartfile) locstart(startfile);
    else  dispwin(0);

restart:

    signal (SIGTSTP, onctlz);
    if (!arglist) printpath(fullpath,True);
    countarg = 0;
startup:
    while ((command = getchar ()) != 'q')
    {

#ifdef _AIX
       showscreen();	/* Workaround for curses library bug in older  */
			/* AIX versions on IBM systems; it is harmless */
			/* to leave it here even after the bug is fixed. */
#endif /* _AIX */

	if (blurb)
	{
	    telluser ("", "");
	    blurb = 0;
	}
	if (ISDIGIT(command)) getcountarg();

       /* GBG fix for cursor keys on DEC and Sun SPARC keyboards */
       /* Change dired.hlp to reflect new keys */
        if ((command == ESC) &&
	    (((i = getchar()) == ESC) || (i == 'O') || (i == '[')))
	{
	    switch (getchar())
	    {
	    case ESC: break;			/* Two escapes mapped to one */
	    case 'x':				/* fall through */
	    case 'A': command = 'k'; break;	/* up arrow */
	    case 'r':				/* fall through */
	    case 'B': command = 'j'; break;	/* down arrow */
	    case 'v':				/* fall through */
	    case 'C': command = 'f'; break;	/* right arrow */
	    case 't':				/* fall through */
	    case 'D': command = 'b'; break;	/* left arrow */
	    case 's':				/* fall through */
	    case 'k': command = 'f'; break;	/* PageDown */
	    case 'y': command = 'b'; break;	/* PageUp */
	    case '5': command = 'b'; (void)getchar() ; break;
					/* ESC [ 5 ~ (Sun type 5) = PageUp  */
	    case '6': command = 'f'; (void)getchar() ; break;
					/* ESC [ 6 ~ (Sun type 5) = PageUp  */
	    }
        }

	switch (command)
	{
	    case ESC: break;
	    case ' ':
		       if (splitflg) prtsplitstr(""); /* clear lower screen */
                       if (!arglist) printpath(fullpath,True);
		       break;

	    case 'a': 		/* Abort completely */
		curxy(0,Tscrlen);
		ceol (0,Tscrlen);
		unsetdpy ();
		exit (1);
		break;
	    case 'Q':    /* abort all the direds and pushd to the
			    dir of the current file */
	       { char bldbuf[MAXLBUFF];

	           bldbuf[0]='\0';
	           bldnam(bldbuf, numarg,curfile,argv);
	           cp=strrchr(bldbuf,'/');
		   curxy(0,Tscrlen);
		   ceol (0,Tscrlen);
		   unsetdpy ();
		   if (cp) {
		      *cp='\0';
		      typein("pushd ");
		      typein(bldbuf);
		      typein("\n");
	           }
		   exit(ABORTSTAT);
	       }
	    case 'A':           /* Abort the chain of direds  ***crh */
		curxy(0,Tscrlen);
		ceol (0,Tscrlen);
		unsetdpy ();
		exit(ABORTSTAT);
		break;
	    case 'P': 		/* Print a file */
		if (file[curfile].ltype == 'd')
		{
		    telluser ("?Can only print files", "");
		    break;
		}
		tmpbuf[0] = '\0';
		bldnam(tmpbuf, numarg, curfile, argv);
		telluser ("Printing...", "");
		while ((i = fork ()) == -1)
		    sleep (3);
		if (i == 0) {
		    	execlp ("print", "print", tmpbuf, 0);
		    telluser ("?Can't find program to list file.\n", "");
		    _exit(1);
		}
		wait (&status);
		break;

	    case 'p':
	       { char bldbuf[MAXLBUFF];
	       strcpy(bldbuf,FNAME(curfile));
	       expandpath(bldbuf);
	       printpath(bldbuf,False);
	       }
	       break;


	    case '/':
	       {
		  char buff[MAXLBUFF];
		  telluser(" ", "");
		  curxy(0,Tscrlen);
		  putchar('/');
		  if (!getline(buff)){
		    telluser("?Pattern too long", "");
		    break;
		  }
		  if (((int)strlen(buff) > 0) && (re_comp(buff) != (char*)NULL))
		  {
		    telluser("?Pattern is not a valid regular expression", "");
		    break;
		  }
		  findfile(1);
		  break;
	       }

	    case '\\' :  /* toggle split screen */

		     if (splitflg) scrlen = Tscrlen - 1;
		     else {
		       if (asswin != -1) scrlen = asswin;
		       else scrlen = 999;
		     }
		     windsize();
		     topfile = curfile/(scrlen - 1) * (scrlen - 1);
		     blank();
		     showscreen();
		     curline= curfile - topfile;
		     curxy(CURSOR,curline);
		     break;


	    case 'n': findfile(1);
		      break;

	    case 'N': findfile(0);
		      break;


	    case '!': 		/* Execute a system command */
		    case '.':		/* repeat the previous ! command */
		telluser ("", "");
		curxy (0, Tscrlen);
		unsetdpy ();
		printf ("Command: ");
		tmpbuf[0] = 'x';	/* dummy kludge */
			if ((command == '!' ? gets(tmpbuf) : strcpy(tmpbuf, combuf))
					      != NULL && tmpbuf[0] != '\0') {
		    register char *op,		/* old ptr */
				  *np;		/* new ptr */
			    char     bldbuf[MAXLBUFF];

			    strcpy(combuf, tmpbuf);	/* remember the command */
		    bldbuf[0] = '\0';
		    op = tmpbuf;
		    special = 0;
		    while ((cc = *(np = skipto(op, "%#"))) != 0) {
			special++;		/* set flag */
			*np++ = '\0';	     /* zap the %/# and bump past it */
			strcat(bldbuf, op);	/* 1st part */
			if (cc == '%')	/* complete file name */
			    bldnam(bldbuf, numarg, curfile, argv);
			else {		/* Had # sign, trailing comp only */
			    if (numarg <= 2)
				strcat (bldbuf, file[curfile].ln.lname);
			    else
			        strcat (bldbuf, file[curfile].ln.namep);
			}

			op = np;
		    }
		    strcat(bldbuf, op);

		    blank ();
		    if (special) {		/* display expanded command */
			printf( "%s\n", bldbuf);
		    }	/* "system" takes long enuf for him to see it */
		    signal(SIGINT, SIG_DFL);	/* temp kludge here... */
		    signal(SIGQUIT, SIG_DFL);	/* should not use 'system' */
		    signal(SIGTSTP, SIG_DFL);
		    cshsystem (bldbuf);
		    signal(SIGTSTP, onctlz);
		    signal (SIGINT, SIG_IGN);
		    signal (SIGQUIT, SIG_IGN);
		    printf ("\nCR to return...");
		    setdpy ();
		    getchar ();
		    blank ();
		    showscreen ();
		    telluser ("", "");
		}
		else {			/* CR only, or EOF, or error */
	       /*
	        * He changed his mind, skip it.  Since we were in
	        * cooked mode, the CR ending the gets is echoed and
	        * we lost the first line of the display.  So we
	        * re-insert it.
		*/
		    setdpy();
	            if (tmpbuf[0] == 'x')	/* means EOF */
			telluser("", "");
		/* null entry, normal case */
		    else if (AL == 0) {		/* no insert line capability */
		        blank();
			showscreen();
		    }
		    else {			/* be a little sneakier */
			curxy(0, Tscrlen-1);	/* go back to where prompt */
			ceol(0, Tscrlen-1);  	/*  is now, and blank it */
			home();
			insline();		/* make some room */
			pentry(topfile);
			putchar('\n');
		    }
		    curxy(CURSOR, curline);
		}
		break;
              case 'R':  /* re-read the directory */
		     {
		       int num;

		     /* remember the name of the current file */
		       strcpy(tmpbuf,FNAME(curfile));
		     deletefiles (argc,argv);  /* delete marked files */
		     if (numdeleted) goto startup; /* undeleted - restart */
		     totfiles = 0;
		     signal(SIGTSTP,SIG_DFL);
		     curxy(0,Tscrlen);
		     ceol (0,Tscrlen);
		     printf("Reading");
		     getfiles (argc,argv);
                     qsort (file, totfiles, sizeof (struct lbuf), compar);
		     /* try to pos cursor where we were */
		     if (findfname(tmpbuf,&num))
		       dispwin(num);
		     else
		       dispwin(0);
		     goto restart;
		     }


	    case 'r': 		/* Reverse sort */
		curxy (0, Tscrlen);
		printf ("reverse ");
		rflg = -1;

	    case 's': 		/* Normal sort */
		if (command == 's')
		{
		    curxy (0, Tscrlen);
		    rflg = 1;
		}
		printf ("sort by [s,n,r,w,c]: ");
		command = getchar ();
		while ((command == '?') || !((command == 'n') ||
		    (command == 'r') || (command == 'w') || (command == 's') ||
		    (command == 'c') || (command == ESC)))
		{
		    curxy (0, Tscrlen);
		    ceol (0, Tscrlen);
		    if (rflg == -1)
			printf ("reverse ");
		    printf ("sort by size, name, read or write date: ");
		    command = getchar ();
		}

		if (command == ESC) {		/* abort */
		    telluser("", "");
		    curxy(CURSOR, curline);
		    break;
		}

		if (command == 's')
			sortyp = SIZE;
		else if (command == 'w')
			sortyp = WRITE;
		else if (command == 'r')
			sortyp = READ;
		else if (command == 'c')
			sortyp = CREATE;
		else
			sortyp = NAME;
		printf ("%c", command);
		qsort (file, totfiles, sizeof (struct lbuf), compar);
		topfile = 0;
		curfile = 0;
		curline = 0;
		blank ();
		showscreen ();
		curxy (CURSOR, 0);
		break;

	    case 'e': 		/* Edit a file or directory */
	    case 'E':           /* go up a dir */
		if ((file[curfile].ltype == 'd') || (command == 'E')){
		    strcpy (tmpbuf, dirednam);
		    strcat(tmpbuf, " ");
		    catargs(tmpbuf, oldargv);
		}
		else {
		    if (((t = getenv("EDITOR")) != NULL) && ((int)strlen(t) > 0)){
			strcat(strcpy(tmpbuf, t), " ");
		    }
		    else
			strcpy (tmpbuf, "vi ");     /***** ex -> vi crh ***/
		}
	        strcpy(tmpbuf2,FNAME(curfile));
	        expandpath(tmpbuf2);
		if (command == 'E') {
		   /* if we got where we are by going down, then 'E'
		      should just quit    */
		   if (camedown) goto quit;

		   /* otherwise form the full path name of the upper node */
		   cp=strrchr(tmpbuf2,'/');
		   if (cp) {
		      *cp = '\0';
		      formstartarg(tmpbuf2);
		      strcat(tmpbuf2,"/..");
		   }
		   else break; /* something would be wrong! */
		   dcanon(tmpbuf2);
		   if (eq(tmpbuf2,fullpath)) {
		      if (eq(tmpbuf2,"/")) telluser("?Already at '/'", "");
		      else telluser("?Can't go up", "");
		      break;
		   }
		   strcat(tmpbuf,startarg);
		}
		else /* 'e' */ {
		  if (file[curfile].ltype == 'd') {
		    if (wasstartfile && eq(startfile,FNAME(curfile)) && quitok)
		       /* we came up this path; just quit */
		       goto quit;
		    strcat(tmpbuf, "-d "); /* indicate we're going down */
		  }
		}
		strcat(tmpbuf,tmpbuf2);
		blank ();
		unsetdpy ();
		signal(SIGTSTP, SIG_DFL);
		diredstat =system (tmpbuf);
	        if ((diredstat>>8)==ABORTSTAT) exit(ABORTSTAT);
		signal(SIGTSTP, onctlz);
		setdpy ();
		if (diredstat != 0) {
		  printf("\n type <cr> to continue");
		  getchar();
		}
		blank ();
		showscreen ();
		if (arglist) telluser ("", "");
		else printpath(fullpath,True);
		break;

	    case 'm': 		/* 'more' a file */
		if (file[curfile].ltype == 'd')
		{
		    telluser ("?Can only page thru files", "");
		    break;
		}
		{ /* allow PAGER environment variable to override builtin */
		    char *p = getenv("PAGER");
		    strcpy (tmpbuf, p ? p : MOREPGM);
		    strcat(tmpbuf," ");	/* supply separating space */
		}
		bldnam(tmpbuf, numarg, curfile, argv);
		blank ();
		unsetdpy ();
		signal(SIGTSTP, SIG_DFL);
		system (tmpbuf);
		signal(SIGTSTP, onctlz);
/*		if (!sigint) {	*/
		printf ("\nCR to return...");
/*  		}		*/
		setdpy ();
 	   	getchar ();
		blank ();
		showscreen ();
		telluser ("", "");
		break;

	    case 'T':	/* don't wait at page end */
	    case 't':	/* quickly type the file -- added 5/81, J.Lepreau */
		if (file[curfile].ltype == 'd')
		{
		    telluser ("?Can only type files", "");
		    break;
		}
		tmpbuf[0] = '\0';
		bldnam(tmpbuf, numarg, curfile, argv);
		if (type(tmpbuf, command == 't')) {   /* little t means wait */
		    blank ();
		    showscreen ();
		    telluser("", "");
		}
		curxy (CURSOR, curline);
		break;

	    case 'l': 		/* Refresh screen */
	    case CTRL('l'):	/* added for editor compatibility -fjl */
		blank ();
		showscreen ();
		telluser ("", "");
		break;
	    case 'c': 		/* Refresh current line */
		curxy (0, curline);
		pentry (curfile);
		curxy (CURSOR, curline);
		break;
	    case CTRL('v'):
	    case 'f': 		/* forward window */
		fscreen (countarg);
		break;
				/* wish we could do meta-v */
	    case 'b': 		/* backward window */
		bscreen (countarg);
		break;
	    case CTRL('n'):
	    case '\r':
	    case 'j' :   /***crh***/
	    case '\n': 		/* next file */
		if (curfile == totfiles - 1)
		    telluser ("?At end of files", "");
		else {
                   if (countarg < 1) countarg = 1;
		   curfile = curfile + countarg;
		   if (curfile > totfiles - 1) {
		      curfile = totfiles - 1;
		      dispfile(curfile);
		      shortcountmess();
		      break;
		   }
		   dispfile(curfile);
	         }
		 break;
	    case '^': 		/* previous file */
	    case CTRL('h'):		/* backspace */
	    case 'k':            /****crh****/
	    case '-':
	    case CTRL('p'):
		if (curfile == 0)
		    telluser ("?At start of files", "");
		else {
		    if (countarg < 1) countarg = 1;
		    curfile = curfile - countarg;
		    if (curfile < 0) {
		       curfile=0;
		       dispfile(curfile);
		       shortcountmess();
		       break;
		    }
		    dispfile(curfile);
		}
	        break;

	    /* ^G info as in vi  */
	    case CTRL('G'): sprintf(tmpbuf,"file %d of %d --%d%%--",
	                    curfile + 1,totfiles, (curfile+1)*100/totfiles);
			  telluser("%s",tmpbuf);
			  break;

	    /* line number addressing as in vi */
	    case 'G': if (countarg < 1) dispfile(totfiles-1);
		      else if (countarg > totfiles) {
			dispfile(totfiles - 1);
			shortcountmess();
		      }
		      else dispfile(countarg-1);
		      break;

	    case 'h': 		/* Help */
	    case '?':
		if (type(helpfile, 1)) {	/* wait */
		    blank ();
		    showscreen ();
		    telluser("", "");
		}
		curxy (CURSOR, curline);
		break;
	    case 'd': 		/* delete file */
		    if (file[curfile].deleted != DELETED) {
		        numdeleted++;
		        file[curfile].deleted = DELETED;
		        printf ("D");
		    }
		    if (curfile == totfiles - 1)
		       curxy (CURSOR, curline);  /***crh 1/84 ***/
		    else {
		      if (curline + 1 == scrlen) fscreen (1);
		      downline ();
		    }
		    break;

	    case 'D': 		/* delete all files */
		for (i = 0; i < totfiles; ++i)
		{
		    if (file[i].deleted != DELETED)
		        file[i].deleted = DELETED;
		}
		numdeleted = totfiles;
		blank ();
		showscreen ();
		telluser ("", "");
		break;

	    case '~': 		/* delete editor backups of form "file~" */
		for (i = 0; i < totfiles; ++i)
		{
		    size_t n = strlen(file[i].ln.lname);

		    if ( (file[i].ln.lname[n-1] == '~') &&
			 (file[i].deleted != DELETED) )
		    {
		        numdeleted++;
		        file[i].deleted = DELETED;
		    }
		}
		blank ();
		showscreen ();
		telluser ("", "");
		break;

	    case '#': 		/* delete editor backups of form "#file" */
		for (i = 0; i < totfiles; ++i)
		{
		    if ( (file[i].ln.lname[0] == '#') &&
			 (file[i].deleted != DELETED) )
		    {
		        numdeleted++;
		        file[i].deleted = DELETED;
		    }
		}
		blank ();
		showscreen ();
		telluser ("", "");
		break;

	    case 'U': 		/* undelete all requested deletions */
		for (i = 0; i < totfiles; ++i)
		{
		    if (file[i].deleted == DELETED)
		        file[i].deleted = NOTDELETED;
		}
		numdeleted = 0;
		blank ();
		showscreen ();
		telluser ("", "");
		break;

	    case 'u': 		/* undelete file */
		    if (file[curfile].deleted == DELETED) {
		        numdeleted--;
		        file[curfile].deleted = NOTDELETED;
		        printf (" ");
		    }
		    if (curfile == totfiles - 1)
		       curxy (CURSOR, curline);  /***crh 1/84 ***/
		    else {
		      if (curline + 1 == scrlen) fscreen (1);
		      downline ();
		    }
		    break;

	    case '[':
	    case '(':
	    case '{':
	    case '<':			/* go to start of file list */
		if (countarg <= 0)
		    dispfile(0);
		else
		{
		    if (countarg > 10)
			countarg = 10;
		    dispfile((int)(((long)countarg * (long)totfiles) / 10L));
		}
		break;

	    case ']':
	    case ')':
	    case '}':
	    case '>':			/* go to end of file list */
		if (countarg <= 0)
		    dispfile(totfiles-1);
		else
		{
		    if (countarg > 10)
			countarg = 10;
		    countarg = 10 - countarg;
		    dispfile((int)(((long)countarg * (long)totfiles) / 10L));
		}
		break;

	    case 'J':			/* jump to top mark, pop mark stack */
		if (mark_top > 0)
		    dispfile(mark_stack[--mark_top]);
		else
		    telluser("?Mark stack is empty: maybe push some with M", "");
		break;

	    case 'L':			/* jump to top view, pop view stack */
		if (view_top > 0)
		{
		    dispfile(view_stack[--view_top]);
		    view_top--;		/* undo the view_top++ in dispfile() */
		}
		else
		    telluser("?View stack is empty", "");
		break;

	    case 'M':			/* save mark */
		if (mark_top < MAX_MARKS)
		{
		    mark_stack[mark_top++] = curfile;
		    telluser("Marked...", "");
		}
		else
		    telluser("?Too many marks pushed: pop some with J", "");
		break;

	    default:
		if (command != -1) /* get around ^Z sending EOF? */
		  telluser ("?Unknown command. Type ? or h for help ", "");
		break;
	}
	countarg = 0;
    }
 quit:
    deletefiles (argc,argv);
    if (numdeleted) goto startup;
    curxy(0,Tscrlen);  /*** crh ***/
    ceol (0,Tscrlen);
    unsetdpy();
    return(errcode);
}

/*** new routines crh 8/82 ***/


static void
getcountarg()
{
  countarg = 0;
  while (ISDIGIT(command)) {
    countarg = 10 * countarg + command - (int)'0';
    command = getchar();
  }
}

static void
typein(argv)
/* puts arg chars into input buffer, i.e., fakes typing them in */
const char *argv;
{
#if defined(HAVE_TERMIOS_H)
    (void)fprintf(stderr,"typein(\"%s\") not yet implemented\n",argv);
#else /* NOT defined(HAVE_TERMIOS_H) */
	struct sgttyb stb, stb2;
/**	int pendin = LPENDIN; **/

	ioctl(2, TIOCGETP, &stb);
	stb2 = stb;
	stb.sg_flags &= ~ECHO;
	ioctl(2, TIOCSETN, &stb);
	for (;; argv++) {
		  if (*argv == '\0') break;
/**			else ioctl(2, TIOCSTI, argv); **/
	}
	ioctl(2, TIOCSETN, &stb2);
/**	ioctl(2, TIOCLBIS, &pendin); **/
#endif /* defined(HAVE_TERMIOS_H) */
}


static handler_t
onctlz (sig)
int sig;
{
  curxy (0, Tscrlen);
  ceol (0,Tscrlen);
  unsetdpy();
  signal(SIGTSTP,SIG_DFL);
  kill(0,SIGSTOP); /*re-send ^z */
  signal(SIGTSTP,onctlz);  /* wake up here ... */
  setdpy();
  blank();
  showscreen();
  if (!intype) telluser("", "");
  else {
    curxy (0, Tscrlen);
    ceol (0,Tscrlen);
  }
#if !defined(HAVE_VOID_HANDLER)
  return (0);
#endif
}

static int
match(filenum)
int filenum;
{
    char *strp;
    strp = (file[filenum].lflags & ISARG) ? file[filenum].ln.namep :
	file[filenum].ln.lname;
    return ( re_exec(strp) );
}

static void
dispfile(fnum)
/* move to the indicated file number */
int fnum;
{
  if (view_top < MAX_VIEW)		/* remember the last view shown */
    view_stack[view_top++] = curfile;
  else
    telluser("?Too many views pushed: pop some with L", "");

  /* redraw if file not already on screen */
  if ((fnum< topfile) || (fnum >=topfile+scrlen)) {
     topfile = fnum/(scrlen - 1) * (scrlen - 1);

     /***crh 2/84 get rid of 1 file final windows unless necessary ***/
     if ((topfile == totfiles - 1) && (topfile - scrlen + 1 >= 0))
       topfile = topfile - scrlen + 1;

     blank();
     showscreen();
  }
  curfile=fnum;
  curline= fnum - topfile;
  curxy(CURSOR,curline);
}

static void
findfile (down)
int down;
{
int i,status;
  status=0;
  if (down) {
    for (i=curfile+1; i < totfiles; i++) { /***crh 1/84 ch from <=*/
      status = match(i);
      if (status != 0) break;
    }
  }
  else {
    for (i=curfile-1; i>= 0 ; i--) {
      status = match(i);
      if (status != 0) break;
    }
  }
  if (status == -1) {
     telluser("?Botched regular expression", "");
     return;
  }
  if (status) dispfile(i);
  else telluser("?No match", "");
}

#if defined(HAVE_TERMIOS_H)
#define KILL_CHAR	tty_save.c_cc[VKILL]
#define ERASE_CHAR	tty_save.c_cc[VERASE]
#else /* NOT defined(HAVE_TERMIOS_H) */
#define KILL_CHAR	ioctlb.sg_kill
#define ERASE_CHAR	ioctlb.sg_erase
#endif /* defined(HAVE_TERMIOS_H) */


static int
getline(s)
char s[];
{
 int status,c,i;
 i=0;
 s[0]='\0';
 status=1;
 while ( ((c=getchar()) != EOF) && (c != '\n') && (c != '\r')) {
   if (c == KILL_CHAR) {
      i=0;
      curxy(1,Tscrlen);
      ceol(1,Tscrlen);
      continue;
   }
   if (c != ERASE_CHAR) {
       if (i < scrwid - 1) {
	 s[i++]=c;
         putchar(c);
       }
       else status=0;
   }
   else {
     if (i > 0) {
       i--;
       putchar(c);
       putchar(' '); putchar(c);
     }
   }
 }
 s[i]='\0';
 return(status);
}

static void
windsize ()
{
    if (scrlen == 0)			/* full screen */
	scrlen = Tscrlen;
    else if (scrlen == 999)		/* means split in half */
        scrlen = (Tscrlen - 1) >> 1;
    if (scrlen < 2)
	scrlen = 2;
    if (Tscrlen < scrlen)
	scrlen = Tscrlen;
    splitflg = (Tscrlen > scrlen+1);	/* 1 extra line for separator */
    if (splitflg) {
	Worklen = Tscrlen - (scrlen+1);	/* size of 'working' window */
	Worktop = scrlen + 1;		/* bottom half for now */
    }
    else
	Worklen = 0;

}

static void
locstart(strtfile)
char strtfile[];
{
    int sfilenum;

       if (findfname(strtfile,&sfilenum))
	 dispwin(sfilenum);
       else {
	 /* if we cant find it, we cant use it */
	 wasstartfile=False;
	 dispwin(0);
       }

}

static int
findfname(filename,num)
char filename[];
int  *num;
{
  int i;
       for (i=0; i<totfiles; i++) {/***crh 1/84 ch from <=*/
	 if eq(FNAME(i),filename) {
	   *num = i;
	   return(True);
	 }
       }
       return(False);
}

static void
formstartarg(name)
/* form start file arg */
char *name;
{
    char tmpbuf3[MAXLBUFF];
    char    *cp;

 if (arglist) strcpy(startarg," -F");
 else strcpy(startarg," -f");
 strcpy(tmpbuf3,name);
 dcanon(tmpbuf3);
 cp=strrchr(tmpbuf3,'/');
 if (cp) strcat(startarg,cp+1);
 else strcat(startarg,tmpbuf3);
 strcat(startarg," ");
}

/*
 * dcanon - canonicalize the pathname, removing excess ./ and ../ etc.  */
static void
dcanon(cp)
	char *cp;
{
	register char *p, *sp;
	register int slash;

	/* assume name begins with a '/' */

	for (p = cp; *p; ) {		/* for each component */
		sp = p;			/* save slash address */
		while(*++p == '/')	/* flush extra slashes */
			;
		if (p != ++sp)
			strcpy(sp, p);
		p = sp;			/* save start of component */
		slash = 0;
		while(*++p)		/* find next slash or end of path */
			if (*p == '/') {
				slash = 1;
				*p = 0;
				break;
			}
		if (*sp == '\0')	/* if component is null */
			if (--sp == cp)	/* if path is one char (i.e. /) */
				break;
			else
				*sp = '\0';
		else if (eq(".", sp)) {
			if (slash) {
				strcpy(sp, ++p);
				p = --sp;
			} else if (--sp != cp)
				*sp = '\0';
		} else if (eq("..", sp)) {
			if (--sp != cp)
				while (*--sp != '/')
					;
			if (slash) {
				strcpy(++sp, ++p);
				p = --sp;
			} else if (cp == sp)
				*++sp = '\0';
			else
				*sp = '\0';
		} else if (slash)
			*p = '/';
	}
}


static int	off;
static char	name[BUFSIZ];

#ifndef V4P2
static char	wd[256];
#else
static char	wd[1024];
#endif


static char *
getfullpath(start)
const char *start;
{
    /*  return the full path name */

    static char	temp[500];		/* address returned */

#ifndef V4P2
    char	dot[500];
    char	dotdot[500];
    int		file;
    struct stat	d, dd;
    struct dirent dir;
    int		rdev, rino;

	off	= -1;
        strcpy(dot,start);
	strcpy(dotdot,start);
	strcat(dotdot,"/..");
	stat("/", &d);
	rdev = d.st_dev;
	rino = d.st_ino;
	for (;;) {
		stat(dot, &d);
		if (d.st_ino==rino && d.st_dev==rdev){
			prname();
			return(wd);
		}
		if ((file = open(dotdot,0)) < 0) {
			telluser("?pwd: cannot open ..\n", "");
			exit(1);
		}
		fstat(file, &dd);
		strcat(dotdot,"/..");
		strcat(dot,"/..");
		if(d.st_dev == dd.st_dev) {
			if(d.st_ino == dd.st_ino){
				prname();
				return(wd);
			}
			do
				if (read(file, (char *)&dir, sizeof(dir)) < sizeof(dir)) {
					telluser("?Read error in ..\n", "");
					printf(dir.d_name); /***/
					exit(1);
				}
			while (dir.d_ino != d.st_ino);
		}
		else do {
				if(read(file, (char *)&dir, sizeof(dir)) < sizeof(dir)) {
					telluser("?read error in ..\n", "");
					printf(dir.d_name); /***/
					exit(1);
				}
				strcpy(temp,dot);
				strcat(temp,"/");
				strcat(temp,dir.d_name);
				stat(temp, &dd);
			} while(dd.st_ino != d.st_ino || dd.st_dev != d.st_dev);
		close(file);
                {
	          register i, j;

	          i = -1;
	          while (dir.d_name[++i] != 0);
	          if ((off+i+2) > BUFSIZ-1){
		          prname();
		          return(wd);
	          }
	          for(j=off+1; j>=0; --j)
		          name[j+i+1] = name[j];
	          off=i+off+1;
	          name[i] = '/';
	          for(--i; i>=0; --i)
		          name[i] = dir.d_name[i];
	       }
	 }
#else
  /* Working dir fix by GBG for local unix systems */
  if (getenv("PWD") != (char *)0)
    strcpy(wd,getenv("PWD"));
  else if (getenv("cwd") != (char *)0)
    strcpy(wd,getenv("cwd"));
  else if (!getwd(wd)) {
    telluser(wd, "");
    exit(1);
  }
  if (eq(start,".")) return (wd);
  else {
    strcpy(temp,wd);
    strcat(temp,"/");
    strcat(temp,start);
    return(temp);
  }

#endif
}

#if 0					/* UNUSED */
static void
prname()
{
	strcpy(wd,"/");
	if (off<0)
		off = 0;
	name[off] = '\0';
	strcat(wd,name);
}
#endif

static void
prtsplitstr(str)
const char str[];
{
int i,top,cur_scrl;

      cur_scrl = totfiles - topfile + 1;	/* topfile starts at 0 */
      top = ((cur_scrl < scrlen) ? cur_scrl : scrlen) + 1;
      curxy(0,top);
      if (CD != 0) putpad(CD);
      /*** 1/84 crh - show embedded control chars - assume ASCII ***/
      for (i=0; i< (int)strlen(str); i++) {
	if (str[i] >= ' ') {
	  if (str[i] >= '\177') {
	    putchar('^');
	    putchar('?');
	  }
	  else putchar(str[i]);
	}
	else {
	  putchar('^');
	  putchar(str[i]+'@');
	}
      }
      curxy (CURSOR, curline);
}

static void
printpath(str,printbottom)
char str[];
int printbottom;
{
static char error[] = "path too long - use split screen to display it";

if (printbottom) {
  if ((int)strlen(str) < scrwid) telluser(str, "");
  else if (splitflg) prtsplitstr(str);
  else telluser(error, "");
}
else if (splitflg) prtsplitstr(str);
else if ((int)strlen(str) < scrwid) telluser(str, "");
else telluser(error, "");
}

static void
expandpath(pname)
char pname[];
{
char temp[MAXLBUFF];

  if (pname[0]!= '/') {
     if (arglist)
	strcpy(temp,getfullpath("."));
     else strcpy(temp,fullpath);
     strcat(temp,"/");
     strcat(temp,pname);
     strcpy(pname,temp);
  }
  dcanon(pname);
}


/* end 8/82 routines */

static void
getfiles (argc,argv)
int argc;
char *argv[];
{
   if (argc == 1)
	Readdir (".");
   else {
	struct stat stbuf;
        arglist=1;
	if (argc == 2) {
	    if (stat(argv[1],&stbuf)==-1) {
	      printf ("\nSorry, %s unreadable.\n", argv[1]);
	      unsetdpy ();
	      exit(1);
	    }
	    if ((stbuf.st_mode & S_IFMT) == S_IFDIR) {
	       Readdir (argv[1]);
	       arglist=0;
	    }
	}
	if (arglist)
	{
	    while (--argc > 0)
	    {
		if (totfiles == MAXFILES)
		    overflow();
		if ((totfiles % DIRPBLOCK) == 0)
		    putchar ('.');
		if (gstat(*++argv) == 0) {
		    file[totfiles].ln.namep = *argv;
		    file[totfiles].lflags |= ISARG;
		    totfiles++;
		}
	    }
	}
    }
    if (totfiles == 0) {
		printf("\n?Empty directory\007\n");
	unsetdpy();
	sleep(1);		/* So user can see it - don't worry if less */
	exit(0);
    }

}

static void
deletefiles (argc,argv)
int argc;
char *argv[];
{
  int i;

    if (numdeleted)
    {
	blank ();
	printf ("The following %s marked for deletion:\n",
		(numdeleted == 1) ? "is" : "are");
	typefiles ();
	printf ("\nShall I delete %s? ",
		(numdeleted == 1) ? "this" : "these");
	if ((command = getchar ()) != 'y')
	{
	    blank ();
	    showscreen ();
	    curxy (0, Tscrlen);
	    ceol (0, Tscrlen);
	    curxy (CURSOR, curline);
	    return;
	}
	else
	{
	  int failures;

	    failures=0;
	    printf ("y\n");
	    for (i = 0; i < totfiles; i++)
		if (file[i].deleted == DELETED) {
		    nambuf[0] = '\0';
		    bldnam(nambuf, numarg, i, argv);
		    if (file[i].ltype == 'd')
			rm (nambuf, 0);
		    else
		    	if (unlink (nambuf) < 0) {
			    printf ("Delete of %s failed.\n", nambuf);
			    failures++;
			}
		}
		if (failures) {
		  printf("<CR> to continue");
		  getchar();
		}
	}
    }
    numdeleted = 0;
}

static void
typefiles ()
{
    int     longsiz,
            i,
            j,
            maxperln,
            numout,
            longthis;
    register char *cp;

    longsiz = numout = 0;
    for (i = 0; i < totfiles; i++)
	if (file[i].deleted == DELETED)
	    if (file[i].lflags & ISARG)
	    {
		if ((int)strlen (file[i].ln.namep) > longsiz)
		{
		    longsiz = strlen (file[i].ln.namep);
		}
	    }
	    else
	    {
		if ((int)strlen (file[i].ln.lname) > longsiz)
		{
		    longsiz = strlen (file[i].ln.lname);
		}
	    }
    maxperln = scrwid / (longsiz + 3);
    if (maxperln < 1) maxperln = 1; /* 3/84 */
    for (i = 0; i < totfiles; i++)
	if (file[i].deleted == DELETED)
	{
	    cp = FNAME(i);
	    longthis = strlen(cp);
            for (; *cp ; cp++) {
              if (( *cp < ' ') || (*cp == '\177')) putchar('?');
              else putchar(*cp);
            }
	    numout++;
	    if ((numout % maxperln) == 0)
		putchar ('\n');
	    else if (numout != numdeleted)
		for (j = 0; j < (longsiz + 3 - longthis); j++)
		    putchar (' ');
	}
}

static void
rm (arg, level)
#ifdef V4P2
/*** shortened for 4.2 by using -r option of rm to do the work ***/
#endif
char    arg[];
int	level;
{
#ifndef V4P2
    struct stat buf;
    struct dirent   dirent;
    char    name[100];
    int     d;
#endif

#ifndef V4P2
    if (stat (arg, &buf))
    {
	return;
    }
    if ((buf.st_mode & S_IFMT) == S_IFDIR)
    {
	if (access (arg, 02) < 0)
	{
	    printf ("%s not deleted.\n", arg);
	    return;
	}
	if ((d = open (arg, 0)) < 0)
	{
	    printf ("rm: %s: cannot read\n", arg);
	    return;
	}
	while (read (d, (char *) & dirent, sizeof (dirent)) == sizeof (dirent))
	{
	    if (dirent.d_ino != 0 && !dotname (dirent.d_name))
	    {
		sprintf (name, "%s/%.14s", arg, dirent.d_name);
		rm (name, level + 1);
	    }
	}
	close (d);
	errcode += Rmdir (arg);
	return;
    }

    if (unlink (arg))
    {
	++errcode;
	printf ("%s not deleted.\n", arg);
    }
#else
    strcpy(tmpbuf,"rm -r ");
    strcat(tmpbuf,arg);
    if (system(tmpbuf)>>8) telluser("?Couldn't delete %s",arg);
    return;
#endif
}

#if 0					/* UNUSED */
static int
dotname (s)
char   *s;
{
    if (s[0] == '.')
	if (s[1] == '.')
	    if (s[2] == '\0')
		return (1);
	    else
		return (0);
	else
	    if (s[1] == '\0')
		return (1);
    return (0);
}
#endif

#ifndef V4P2
static int
Rmdir (f)
char   *f;
{
    int     i;
    wait_arg_t status;

    if (dotname (f))
	return (0);
    while ((i = fork ()) == -1)
	sleep (3);
    if (i == 0) {
    	execl ("/bin/rmdir", "rmdir", f, 0);
    	execl ("/usr/bin/rmdir", "rmdir", f, 0);
    	printf ("rm: can't find rmdir\n");
    	_exit(1);
    }
    wait (&status);
    return (status);
}
#endif

static void
shortcountmess()
{
  telluser("?Used shortened count", "");
}

static void
dispwin(filenum)
/* forces a new window */
int filenum;
{
  topfile = -999;  /* force a redraw */
  dispfile(filenum);
}

static void
fscreen (count)
int count;
{
  if (topfile + scrlen - 1 >= totfiles - 1) {  /***crh 1/84 ch from >*/
      telluser ("?No remaining windows", "");
      return;
  }
  if (count <= 1) count=1;
  else if ((topfile + count*(scrlen-1)) > totfiles - 1)
  {
     dispwin(totfiles - 1);
     shortcountmess();
     return;
  }
 dispwin(topfile + count*(scrlen-1));
}

static void
bscreen (count)
int count;
{
  if (topfile - scrlen + 1 < 0) {
      telluser ("?No previous windows", "");
      return;
  }
  if (count <= 1) count=1;
  else if ((topfile -count*(scrlen-1)) < 0) {
    dispwin(0);
    shortcountmess();
    return;
  }
  dispwin(topfile - count*(scrlen-1));
}

static void
showscreen ()
{
    int     i,
	    x1,x2,
            numprint;
    home ();
    setbuf(stdout, bufout);	/* buffered output here, faster! --fjl */
    numprint = 0;
    for (i = topfile; (numprint < scrlen) && (i < totfiles); i++)
    {
	numprint++;
	pentry (i);
	putchar ('\n');
    }
    if (splitflg) {
	strcpy(newdivider,divider);
	x1 = (topfile * LW) / totfiles;
	x2 = ((topfile+scrlen-1)*LW)/totfiles;
	if (x1==x2) newdivider[x1]='o';
	else {
	  if ((x1==0) && (topfile==0)) newdivider[0]='[';
	  else newdivider[x1]='(';
#if 0
	  if (x2>=LW || topfile+scrlen >= totfiles) newdivider[LW]=']';
	  else newdivider[x2]=')';
#else  /* fix off-by-one error */
	  if (x2>=LW || topfile+scrlen >= totfiles) newdivider[LW-1]=']';
	  else newdivider[x2-1]=')';
#endif
	}
    	printf ("%s\n", newdivider);
    }
    fflush(stdout);		/* reset for display functions */
    setbuf(stdout, (char *)NULL);
}

static void
Readdir (dir)			/* Reads directory dir */
const char   *dir;
{
    register int    j;
    struct dirent *dentryp;
    DIR *dirp;

#if defined(HAVE_OPENDIR)
    if ((dirp = opendir(dir)) == NULL)
#else
    FILE *dirf;

    if ((dirf = fopen (dir, "r")) == NULL)
#endif
    {
	printf ("\nSorry, %s unreadable.\n", dir);
	unsetdpy ();
	exit(1);
    }
#if defined(HAVE_OPENDIR)
    while ((dentryp = readdir(dirp)) != NULL)
#else
    for (;;)
#endif
    {
#if !defined(HAVE_OPENDIR)
	struct dirent    dentry;

	if (fread ((char *) & dentry, sizeof (dentry), 1, dirf) != 1)
	    break;
	dentryp = &dentry;
#endif
	if ((dentryp->d_ino == 0)
	    || ((dentryp->d_name[0] == '.') &&
		((dentryp->d_name[1] == '\0') ||
		 ((dentryp->d_name[1] == '.') && (dentryp->d_name[2] == '\0')))))
	    continue;

	if (totfiles == MAXFILES)
	    overflow();			/* abort, too may files */
	/* Just ignore if can't find the file, dir may be changing */
	if (gstat (makename (dir, dentryp->d_name, D_NAMLEN(dentryp))) == 0) {	/* 0 == Ok */
	    file[totfiles].lnum = dentryp->d_ino;
#if defined(HAVE_DIRENT_H)
	    (void)strcpy(file[totfiles].ln.lname,dentryp->d_name);
#else
	    /* old <sys/dir.h> style might lack NUL termination */
	    for (j = 0; j < D_NAMLEN(dentryp); j++)
	        file[totfiles].ln.lname[j] = dentryp->d_name[j];
	    file[totfiles].ln.lname[j] = '\0'; /* ensure NUL termination */
#endif
	    totfiles++;
	    if (totfiles % 10 == 0)
	        putchar ('.');
	}
    }
#if defined(HAVE_OPENDIR)
    closedir (dirp);
#else
    fclose (dirf);
#endif
}

static int
gstat (gname)			/* Stats the file with name */
const char   *gname;
{

    file[totfiles].lflags = 0;
    file[totfiles].lnum = 0;
    file[totfiles].ltype = '-';

    if (stat (gname, &statbuf) < 0)
    {
	return(-1);
    }
    file[totfiles].lnum = statbuf.st_ino;
    file[totfiles].lsize = statbuf.st_size;
    switch (statbuf.st_mode & S_IFMT)
    {
	case S_IFDIR:
	    file[totfiles].ltype = 'd';
	    break;
	case S_IFBLK:
	    file[totfiles] .ltype = 'b';
	    file[totfiles].lsize = statbuf.st_rdev;
	    break;
	case S_IFCHR:
	    file[totfiles].ltype = 'c';
	    file[totfiles].lsize = statbuf.st_rdev;
	    break;
    }
    file[totfiles].deleted = NOTDELETED;
    file[totfiles].lflags = statbuf.st_mode & ~S_IFMT;
    file[totfiles].luid = statbuf.st_uid;
    file[totfiles].lgid = statbuf.st_gid;
    file[totfiles].lnl = statbuf.st_nlink;
    file[totfiles].latime = statbuf.st_atime;
    file[totfiles].lctime = statbuf.st_ctime;
    file[totfiles].lmtime = statbuf.st_mtime;
    totblocks += nblock (statbuf.st_size);
    return(0);
}

static char *
makename (dir, filen, len)
const char   *dir,
       *filen;				/* NB: might NOT be NUL terminated */
int len;
{
    static char dfile[MAXFILEPATHLEN];
    register char  *dp;
    register const char *fp;
    register int    i;

    dp = dfile;
    fp = dir;
    while (*fp)
	*dp++ = *fp++;
    *dp++ = '/';
    fp = filen;
    for (i = 0; i < len; i++){
	*dp++ = *fp++;
	if (!*fp) break;
    }
    *dp = 0;				/* ensure NUL termination */
    return (dfile);
}

static long
nblock (size)
long    size;
{
    return ((size + 511) >> 9);
}

static void
pentry (whichone)
int     whichone;
{
    register    i,t;
    register char  *cp;

    if (file[whichone].lnum == (unsigned long)(-1))
	return;
    if (iflg)
	printf ("%5lu ", (unsigned long)file[whichone].lnum);
    if (sflg)
	printf ("%4ld ", nblock (file[whichone].lsize));
    if (lflg)
    {
	putchar (file[whichone].ltype);
	pmode (file[whichone].lflags);
	printf ("%2lu ", (unsigned long)file[whichone].lnl);
	t = file[whichone].luid;
	if (gflg)
	    t = file[whichone].lgid;
	if (getname (t, userbuf) == 0)
	    printf ("%-14.14s", userbuf);
	else
	    printf ("%-14d", t);
	if (file[whichone].ltype == 'b' || file[whichone].ltype == 'c')
	    printf ("%3ld,%3ld", (long)major ((int) file[whichone].lsize),
		    (long)minor ((int) file[whichone].lsize));
	else
	    printf ("%7ld", file[whichone].lsize);
	if ((sortyp == WRITE) || (sortyp == NAME) || (sortyp == SIZE))
	{
	    cp = ctime (&file[whichone].lmtime);
	    if (file[whichone].lmtime < year)
		printf (" %-7.7s %-4.4s ", cp + 4, cp + 20);
	    else
		printf (" %-12.12s ", cp + 4);
	}
	else if (sortyp == READ)
	{
	    cp = ctime (&file[whichone].latime);
	    if (file[whichone].latime < year)
		printf (" %-7.7s %-4.4s ", cp + 4, cp + 20);
	    else
		printf (" %-12.12s ", cp + 4);
	}
	else if (sortyp == CREATE)
	{
	    cp = ctime (&file[whichone].lctime);
	    if (file[whichone].lctime < year)
		printf (" %-7.7s %-4.4s ", cp + 4, cp + 20);
	    else
		printf (" %-12.12s ", cp + 4);
	}
    }
    printf ("%c", file[whichone].deleted ? 'D' : ' ');

    if (file[whichone].lflags & ISARG)
	cp = file[whichone].ln.namep;
    else
	cp = file[whichone].ln.lname;
    for (i = 0; ( *cp && (i < 29)); i++) {
      if (( *cp < ' ') || (*cp == '\177')) putchar('?');
      else putchar(*cp);
      cp++;
    }
}

static int
getname (uid, buf)
int     uid;
char    buf[];
{
    int     j,
            c,
            n,
            i;

    if (uid == lastuid)
	return (0);
    if (pwdf == NULL)
	return (-1);
    rewind (pwdf);
    lastuid = -1;
    do
    {
	i = 0;
	j = 0;
	n = 0;
	while ((c = fgetc (pwdf)) != '\n')
	{
	    if (c == EOF)
		return (-1);
	    if (c == ':')
	    {
		j++;
		c = '0';
	    }
	    if (j == 0)
		buf[i++] = c;
	    if (j == 2)
		n = n * 10 + c - '0';
	}
    } while (n != uid);
    buf[i++] = '\0';
    lastuid = uid;
    return (0);
}

static int     m1[] =
{
    1, S_IREAD >> 0, 'r', '-'
};
static int     m2[] =
{
    1, S_IWRITE >> 0, 'w', '-'
};
static int     m3[] =
{
    2, S_ISUID, 's', S_IEXEC >> 0, 'x', '-'
};
static int     m4[] =
{
    1, S_IREAD >> 3, 'r', '-'
};
static int     m5[] =
{
    1, S_IWRITE >> 3, 'w', '-'
};
static int     m6[] =
{
    2, S_ISGID, 's', S_IEXEC >> 3, 'x', '-'
};
static int     m7[] =
{
    1, S_IREAD >> 6, 'r', '-'
};
static int     m8[] =
{
    1, S_IWRITE >> 6, 'w', '-'
};
static int     m9[] =
{
    2, S_ISVTX, 't', S_IEXEC >> 6, 'x', '-'
};

static int    *m[] =
{
    m1, m2, m3, m4, m5, m6, m7, m8, m9
};

static void
pmode (aflag)
int aflag;
{
    register int  **mp;

    flags = aflag;
    for (mp = &m[0]; mp < &m[sizeof (m) / sizeof (m[0])];)
	Select (*mp++);
}

static void
Select (pairp)
register int   *pairp;
{
    register int    n;

    n = *pairp++;
    while (--n >= 0 && (flags & *pairp++) == 0)
	pairp++;
    putchar (*pairp);
}

static int
compar (pp1, pp2)
const void *pp1;
const void *pp2;
{
    register const struct lbuf   *p1,
                           *p2;

    p1 = (const struct lbuf*)pp1;
    p2 = (const struct lbuf*)pp2;
    if (p1 -> lflags & ISARG && p1 -> ltype == 'd')
    {
	if (!(p2 -> lflags & ISARG && p2 -> ltype == 'd'))
	    return (1);
    }
    else
    {
	if (p2 -> lflags & ISARG && p2 -> ltype == 'd')
	    return (-1);
    }
    if (sortyp == SIZE)
    {
	if (p2 -> lsize == p1 -> lsize)
	    return (0);
	if (p2 -> lsize > p1 -> lsize)
	    return (rflg);
	return (-rflg);
    }
    else if (sortyp == WRITE)
    {
	if (p2 -> lmtime == p1 -> lmtime)
	    return (0);
	if (p2 -> lmtime > p1 -> lmtime)
	    return (rflg);
	return (-rflg);
    }
    else if (sortyp == READ)
    {
	if (p2 -> latime == p1 -> latime)
		return (0);
	if (p2 -> latime > p1 -> latime)
		return (rflg);
	return (-rflg);
    }
    else if (sortyp == CREATE)
    {
	if (p2 -> lctime == p1 -> lctime)
		return (0);
	if (p2 -> lctime > p1 -> lctime)
		return (rflg);
	return (-rflg);
    }
    return (rflg * strcmp (p1 -> lflags & ISARG ? p1 -> ln.namep : p1 -> ln.lname,
		p2 -> lflags & ISARG ? p2 -> ln.namep : p2 -> ln.lname));
}

#if 0					/* UNUSED */
static void
ceod()
{
}
#endif

static void
ceol (x,y)
int x,y;
{
    int i;
    if (CE != 0) putpad (CE);
    else {  /***crh for terms w/o ce, put blanks & restore cursor **/
      for (i=x; i <  scrwid ; i++ ) putchar(' ');
      curxy(x,y);
    }
}

static void
blank ()
{
    putpad (CL);
}

static void
home ()
{
     curxy(0,0);   /* crh putpad (HO); */
}

static void
insline ()
{
    putpad (AL);
}

/* Yes, folks, we use direct cursor addressing to get to next line!
   Before you mumble "What sort of cretin would do this?" here's
   the reason. We don't use \n since that obviously won't work.
   We don't use \012 since virgin version 7 makes that into a crlf.
   We don't use raw mode since we type out help files efficently,
   and we don't want to switch modes all the time. So enjoy. -- SMC */

static void
downline ()
{
    curxy (CURSOR, ++curline);
    curfile++;
}

#if 0					/* UNUSED */
static void
upline ()
{
    putpad (UP);
    curline--;
    curfile--;
}
#endif

/*VARARGS1*/
static void
telluser (msg, arg)
const char   *msg;
const char   *arg;
{
    curxy (0, Tscrlen);
    ceol (0,Tscrlen);
    printf (msg, arg);
    curxy (CURSOR, curline);
    blurb++;
}

static void
curxy (col, lin)
int col;
int lin;
{
    putpad (tgoto (CM, col, lin));
}


static int
nottext(fn)
const char fn[];
{
/** routine to try to determine whether the file is obviously not a text
file. Adapted from "more".   *** crh ***/

int c;
FILE *f;

if ((f=fopen(fn, "r")) == NULL) return(False);
c = getc(f);

/* Try to see whether it is an ASCII file */

switch ((c | *f->_ptr << 8) & 0177777) {
    case 0405:
    case 0407:
    case 0410:
    case 0411:
    case 0413:
    case 0177545:
	fclose (f);
	return (True);
    default:
	fclose (f);
	return(False);
    }
 }

static int
Getch(fd)
FILE    *fd;
{
  if (wasspecial) {
     wasspecial = False;
     return (specialcc);
  }
  else return (getc(fd));
}

static int
type (filestr, waitflg)    /* Modified to type help file & others. fjl 5/81 */
const char *filestr;		/* Kludgy now with split screen stuff! */
int waitflg;
{
    int     helpfd = 5;
    FILE    *fd = stdin;
    register int     i,j, n;
    register int     cc = 0;
    int	    cur_scrl;		/* current screen length */
    char    helpbuf[512];

    if (nottext(filestr)) {
       telluser("?Not a text file:%s", filestr);
       return (False);
    }
    if (!splitflg)
	helpfd = open(filestr, 0);
    else
    	fd = fopen(filestr, "r");
    if (helpfd < 0 || fd == NULL) {
    	telluser("?Unable to open %sfile",strcmp(filestr,helpfile)?"":"help ");
	return (False);
    }

    signal(SIGINT, catchint);
    sigint = 0;
    intype=1;

    if (!splitflg) {
    	blank();
    	while ((i = read (helpfd, helpbuf, 512)) > 0 && !sigint)
	    write (1, helpbuf, i);
	close(helpfd);
    }
    else {
	setbuf(stdout, bufout);		/* to speed it up */
	cur_scrl = totfiles - topfile + 1;	/* topfile starts at 0 */
	Worktop = ((cur_scrl < scrlen) ? cur_scrl : scrlen) + 1;
	if (!splitflg) {
	   blank();
	   Worktop=0;
	}
	do {
	    curxy(0, Worktop);
	    /**** crh  replace call to ceod()  ****/
	    if (CD != 0)  putpad(CD);
	    else if (!splitflg) blank();
	    wasspecial = False;
	    for (i = Worktop; (i < Tscrlen) && !sigint && (cc != EOF); i++) {
		n = 0;
		while ((cc = Getch(fd)) != EOF) {
		    if (cc == '\t') n= (n/8) * 8 + 8;
chklen:
		    if (n >= scrwid  /*+1 */ ) { /* Use only 79 cols crh***/
			if (cc != '\n') {
			    if (!wasspecial) ungetc(cc, fd);
			       else specialcc=cc;
			    cc = '\n';
			}
		    }
		    if (cc >= 127) {
			putchar('^');
			cc = '?';
			n++;
			wasspecial = True;
			goto chklen;
		    }
		    else
		    if ( (cc>=0) && (cc<=31) && (cc != '\t') && (cc != '\n') ) {
			 /* form the graphic of the control char */
			putchar('^');
			cc = cc + 64;
			n++;
			wasspecial = True;
			goto chklen;
		    }
		    if (cc != '\n') {
		       putchar(cc);
		       wasspecial = False;
		    }
		    else {
		       if ((CD==0) & splitflg)
			  ceol(n,i);
		       putchar('\n');
		       break;
		    }
		    if (cc != '\t') n++;
		}
	    }
	       if (CD==0 && splitflg)
	         for (j= i; j < Tscrlen;j++ ) {
		    curxy(0, j);
		    ceol (0,j);
	         }
	    fflush(stdout);
	} while (!sigint && (cc != EOF) && waitchk(waitflg));

	if (feof(fd))
	    printf("===== End-of-File =====\n");
	fflush(stdout);
	setbuf(stdout, (char *)NULL);
	fclose(fd);
	return(False);	/* needs no re-display */
    }

    if (!splitflg) {		/* redundant now... */
    	if (!sigint) {
	    curxy(0, Tscrlen);
	    printf ("CR to return...");
    	    getchar ();
    	}
    	blank ();
	return (True);
    }
    return(False);
}

static int
waitchk(waitflg)
int waitflg;
{
	if (!waitflg)
	    return(1);
	fflush(stdout);
	setbuf(stdout, (char *)NULL);
	curxy(0, Tscrlen);
	ceol(0,Tscrlen);
	printf ("---Continue---");
	curxy(0, Tscrlen);
		if (getchar() == 'q')
		    sigint = 1;		/* simulate interrupt */
	ceol(0,Tscrlen);
	setbuf(stdout, bufout);
	if (sigint)
	    return(0);		/* avoids clear of screen */
	return(1);
}

static void
setdpy ()
{
#if defined(HAVE_TERMIOS_H)
    struct termios tty;		/* XPG3, POSIX.1, FIPS 151-1 interface */

    if (tcgetattr(fileno(stdin),&tty) != -1)
    {
	tty_save = tty;
	tty.c_iflag &= ~(INLCR | ICRNL | ISTRIP | IXON | BRKINT);

#if defined(IUCLC)
	tty.c_iflag &= ~IUCLC;		/* absent from POSIX */
#endif /* defined(IUCLC) */

	tty.c_lflag &= ~(ECHO | ICANON);
	tty.c_cc[VMIN] = 5;		/* MIN */
	tty.c_cc[VTIME] = 2;		/* TIME */
	(void)tcsetattr(fileno(stdin),TCSANOW,&tty);
    }
#else /* NOT defined(HAVE_TERMIOS_H) */
    ioctl (0, TIOCGETP, &ioctlb);
    ioctlb.sg_flags |= CBREAK;
    ioctlb.sg_flags &= ~ECHO;
    ioctl (0, TIOCSETP, &ioctlb);
#endif /* defined(HAVE_TERMIOS_H) */
}

static void
unsetdpy ()
{
#if defined(HAVE_TERMIOS_H)
    (void)tcsetattr(fileno(stdin),TCSANOW,&tty_save);
#else /* NOT defined(HAVE_TERMIOS_H) */
    ioctlb.sg_flags &= ~CBREAK;
    ioctlb.sg_flags |= ECHO;
    ioctl (0, TIOCSETP, &ioctlb);
#endif /* defined(HAVE_TERMIOS_H) */
}

static void
Getcap ()
{
    char   *ap;
    char   *term;
    char   *xPC;

    term = getenv ("TERM");
    if (term == 0)
    {
	fprintf (stderr, "No TERM in environment\n");
	exit(1);
    }

    switch ((int)(tgetent (tbuf, term)))
    {
	case -1:
	    fprintf (stderr, "Cannot open termcap file\n");
	    exit (2);
	case 0:
	    fprintf (stderr, "%s: unknown terminal", term);
	    exit (3);
    }

    ap = tcapbuf;

    Tscrlen = tgetnum ("li") - 1;
    scrwid = tgetnum ("co") - 1; /* lose 1 so won't scroll in last line */

#ifdef TIOCGWINSZ
    /* On SunOS, tgetnum("li") returns the true window size, but under
       ULTRIX, it returns the value from the termcap entry, which is usually
       wrong in a window environment.  Get the true values if possible. */
    {
	struct winsize win;

	if (ioctl (0, TIOCGWINSZ, &win) == 0)
	{
		if (win.ws_col)
			scrwid = win.ws_col;
		if (win.ws_row)
			Tscrlen = win.ws_row;
	}
    }
#endif

    if (scrwid > MAXWIDTH)
	scrwid = MAXWIDTH;

#ifdef ultrix
    scrwid--;				/* ULTRIX wraps in last column */
    Tscrlen--;				/* ULTRIX scrolls when cursor */
					/* enters last line in */
					/* full-window mode, so */
					/* shorten the height by 1 line */
#endif


    UP = tgetstr ("up", &ap);
    CD = tgetstr ("cd", &ap);
    CE = tgetstr ("ce", &ap);
    HO = tgetstr ("ho", &ap);
    CL = tgetstr ("cl", &ap);
    CM = tgetstr ("cm", &ap);
    AL = tgetstr ("al", &ap);		/* insert line, optional */

    xPC = tgetstr ("pc", &ap);
    if (xPC)
	PC = *xPC;

    /*** crh changed error test ***/
    if ((CM == 0) || (CL == 0) || (UP == 0)  )
    {
	fprintf (stderr, "Tty must have cursor addr, clear, and 4 cursor motions.\n");
	exit (1);
    }
    if (Tscrlen <= 0 || scrwid <= 0)
    {
	fprintf (stderr, "Must know the screen size\n");
	exit (1);
    }
}

#if STDC
#if defined(__sgi)
static int
outch (int c)
#else
static int
outch (char c)
#endif
#else
static int
outch (c)
int c;
#endif
{
    return (putchar (c));

}

static void
putpad (str)
char   *str;
{
    if (str)
	tputs (str, 0, outch);
}

static handler_t
catchint(sig)
int sig;
{
	signal(SIGINT, SIG_IGN);	/* reset it */
	sigint = 1;
#if !defined(HAVE_VOID_HANDLER)
	return (0);
#endif
}

static char *
bldnam(str, narg, filidx, argv)
char *str;
int narg;
int filidx;
char *argv[];
{

	if (narg == 1)
	    strcat (str, file[filidx].ln.lname);
	else
	    if (narg == 2) {
		if (arglist)
		   strcat (str, file[filidx].ln.namep);
		else {
		   strcat (str, argv[1]);
		   strcat (str, "/");
		   strcat (str, file[filidx].ln.lname);
		}
	    }
	    else
		strcat (str, file[filidx].ln.namep);
	return(str);
}

static char *
catargs(str, argv)
char *str;
char *argv[];
{
	char *cp;

	for (++argv; *argv; argv++) {
		if (**argv == '-') {
		  cp = &(**argv) + 1;
		  if (*cp != 'f' && *cp != 'F' && *cp != 'd') {
			strcat(str, " ");
			strcat(str, *argv);
		  }
		}
	}
	return(str);
}

static void
overflow()
{
    printf("\n?Too many files\007\n");
    unsetdpy();
    sleep(1);		/* So user can see it - don't worry if less */
    exit(1);
}

static char *
skipto (string,charset)
const char *string;
const char *charset;
{
	register const char *setp,*strp;
	register int found;

	found = 0;			/* not found yet */
	strp = string;			/* start at first char */

	while (*strp && !found) {	/* until null or found */
		/* find first char in charset matching *strp */
		for (setp=charset; (*setp) && (*setp != *strp); setp++) ;
		if (*setp)	found = 1;	/* matches a char */
		else		strp++;		/* else keep looking */
	}

	return ((char*)strp);
}
